;-----------------------------------------------------------------------------------
; BareNumbersOS v.0.4 - GPL3 - Apr 2013
;
;-----------------------------------------------------------------------------------
; IDE Hard Drive Functions
;-----------------------------------------------------------------------------------

align 16


; NOTE:	These functions use LBA48. Maximum visible drive size is 128PiB
; -----------------------------------------------------------------------------
; IDEreadsectorsLBA48 -- Read sectors on the hard drive
; IN:	RAX = starting sector to read
;	RCX = number of sectors to read (1 - 256)
;	RDI = memory location to store sectors
;	RBX = hdbase
;	RDX = hdid
; OUT:	RAX = RAX + number of sectors that were read
;	RCX = number of sectors that were read (0 on error)
;	RDI = RDI + (number of sectors * 512)
;	All other registers preserved
IDEreadsectorsLBA48:
	push r10
	push r9
	push r8
	push rdx
	push rcx
	push rbx
	push rax

	mov r9, rbx		; Store hdbase
	mov r10, rdx		; Store hdid
		
	push rcx		; Save RCX for use in the read loop
	mov rbx, rcx		; Store number of sectors to read
	cmp rcx, 256
	ja .readsectors_fail	; Over 256? Fail!
	jne .readsectors_skip	; Not 256? No need to modify CL
	xor rcx, rcx		; 0 translates to 256
.readsectors_skip:

	push rax		; Save RAX since we are about to overwrite it
	mov dx, r9w		; e.g.: 0x01F0 [hdbase]

	add dx, 2		; 0x01F2
   	xchg ch, cl
	mov al, cl		; Sector count Port 15:8 (Read CH sectors = 0)
	out dx, al
   	xchg ch, cl
	mov al, cl		; Sector count Port 7:0	(Read CL sectors)
	out dx, al

	pop rax			; Restore RAX which is our sector number
	mov r8, rax		; Store the value
;------
	inc dx			; 0x01F3
	shr rax, 24
	out dx, al		; LBA Low Port 31:24
	mov rax, r8
	out dx, al		; LBA Low Port 7:0

	inc dx			; 0x01F4
	mov rax, r8
	shr rax, 32		; LBA Mid Port 39:32
	out dx, al
	mov rax, r8
	shr rax, 8		; LBA Mid Port 15:8
	out dx, al

	inc dx			; 0x01F5
	mov rax, r8
	shr rax, 40		; LBA High Port 47:40
	out dx, al
	mov rax, r8
	shr rax, 16		; LBA High Port 23:16
	out dx, al

	inc dx			; 0x01F6 - Device Port. Bit 6 set for LBA mode, Bit 4 for device (0 = master, 1 = slave)
	xor rax, rax	 	; Clear bits
	or al, 01000000b	; Turn bit 6 on since we want to use LBA addressing
	or al, r10b		; Set bit 4 for device [hdid]
	out dx, al

	inc dx			; 0x01F7 - Command Port
	mov al, 0x24		; Read sector(s). 0x24 if LBA48
	out dx, al

	mov rcx, 4
.readsectors_wait:
	in al, dx		; Read status from 0x01F7
	test al, 0x80		; BSY flag set?
	jne .readsectors_retry
	test al, 0x08		; DRQ set?
	jne .readsectors_dataready
.readsectors_retry:
	dec rcx
	ja .readsectors_wait
.readsectors_nextsector:
	in al, dx		; Read status from 0x01F7
	test al, 0x80		; BSY flag set?
	jne .readsectors_nextsector
	test al, 0x21		; ERR or DF set?
	jne .readsectors_fail

.readsectors_dataready:
	sub dx, 7		; Data port (0x1F0)
	mov rcx, 256		; Read 
	rep insw		; Copy a 512 byte sector to RDI
	add dx, 7		; Set DX back to status register (0x01F7)
	in al, dx		; Delay ~400ns to allow drive to set new values of BSY and DRQ
	in al, dx
	in al, dx
	in al, dx

.CheckForBusy:
        NOP
        in al, dx
        test al , 0x80		; BSY should be cleared
        jnz .CheckForBusy

	dec rbx			; RBX is the "sectors to read" counter
	cmp rbx, 0
	jne .readsectors_nextsector

	jmp .readsectors_ok


.readsectors_fail:
	pop rcx

	pop rax
	pop rbx
	pop rcx
	pop rdx
	pop r8
	pop r9
	pop r10
	xor rcx, rcx		; Set RCX to 0 since nothing was read
	jmp .end
.readsectors_ok:
	pop rcx

	pop rax
	pop rbx
	add rax, rcx
	pop rcx
	pop rdx
	pop r8
	pop r9
	pop r10
.end:

ret

; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; IDEwritesectorsLBA48 -- Write sectors on the hard drive
; IN:	RAX = starting sector to write
;	RCX = number of sectors to write (1 - 256)
;	RSI = memory location of sectors
;	RBX = hdbase
;	RDX = hdid
; OUT:	RAX = RAX + number of sectors that were written
;	RCX = number of sectors that were written (0 on error)
;	RSI = RSI + (number of sectors * 512)
;	All other registers preserved
IDEwritesectorsLBA48:
	push r10
	push r9
	push r8
	push rdx
	push rcx
	push rbx
	push rax

	mov r9, rbx		; Store hdbase
	mov r10, rdx		; Store hdid

	push rcx		; Save RCX for use in the write loop
	mov rbx, rcx		; Store number of sectors to write
	cmp rcx, 256
	ja .writesectors_fail	; Over 256? Fail!
	jne .writesectors_skip	; Not 256? No need to modify CL
	xor rcx, rcx		; 0 translates to 256
.writesectors_skip:	

	push rax		; Save RAX since we are about to overwrite it
	mov dx, r9w		; e.g.: 0x01F0 [hdbase]

	add dx, 2		; 0x01F2
   	xchg ch, cl
	mov al, cl		; Sector count Port 15:8 (Read CH sectors = 0)
	out dx, al
   	xchg ch, cl
	mov al, cl		; Sector count Port 7:0	(Read CL sectors)
	out dx, al

	pop rax			; Restore RAX which is our sector number
	mov r8, rax		; Store the value
;------
	inc dx			; 0x01F3
	shr rax, 24
	out dx, al		; LBA Low Port 31:24
	mov rax, r8
	out dx, al		; LBA Low Port 7:0

	inc dx			; 0x01F4
	mov rax, r8
	shr rax, 32		; LBA Mid Port 39:32
	out dx, al
	mov rax, r8
	shr rax, 8		; LBA Mid Port 15:8
	out dx, al

	inc dx			; 0x01F5
	mov rax, r8
	shr rax, 40		; LBA High Port 47:40
	out dx, al
	mov rax, r8
	shr rax, 16		; LBA High Port 23:16
	out dx, al

	inc dx			; 0x01F6 - Device Port. Bit 6 set for LBA mode, Bit 4 for device (0 = master, 1 = slave)
	xor rax, rax	 	; Clear bits
	or al, 01000000b	; Turn bit 6 on since we want to use LBA addressing
	or al, r10b		; Set bit 4 for device [hdid]
	out dx, al

	inc dx			; 0x01F7 - Command Port
	mov al, 0x34		; Write sector(s). 0x34 if LBA48
	out dx, al

	mov rcx, 4
.writesectors_wait:
	in al, dx		; Read status from 0x01F7
	test al, 0x80		; BSY flag set?
	jne .writesectors_retry
	test al, 0x08		; DRQ set?
	jne .writesectors_dataready
.writesectors_retry:
	dec rcx
	ja .writesectors_wait
.writesectors_nextsector:
	in al, dx		; Read status from 0x01F7
	test al, 0x80		; BSY flag set?
	jne .writesectors_nextsector
	test al, 0x21		; ERR or DF set?
	jne .writesectors_fail

.writesectors_dataready:
	sub dx, 7		; Data port (0x01F0)
	mov rcx, 256		; Write 256 words (512 bytes)
.writesectors_nextword:
	outsw			; Cannot use rep as a small delay is needed between each out
	sub rcx, 1
	cmp rcx, 0
	jne .writesectors_nextword
	add dx, 7		; Set DX back to Command / Status Register (0x01F7)
	mov al, 0xE7		; Cache Flush command
	out dx, al
	in al, dx		; Delay ~400ns to allow drive to set new values of BSY and DRQ
	in al, dx
	in al, dx
	in al, dx

.CheckForBusy:
        NOP
        in al, dx
        test al , 0x80		; BSY should be cleared
        jnz .CheckForBusy

	dec rbx			; RBX is the "sectors to write" counter
	cmp rbx, 0
	jne .writesectors_nextsector

	jmp .writesectors_ok	

.writesectors_fail:
	pop rcx
	pop rax
	pop rbx
	pop rcx
	pop rdx
	pop r8
	pop r9
	pop r10
	xor rcx, rcx		; Set RCX to 0 since nothing was written
	jmp .end
.writesectors_ok:
	pop rcx
	pop rax
	pop rbx
	add rax, rcx
	pop rcx
	pop rdx
	pop r8
	pop r9
	pop r10

.end:

ret

; -----------------------------------------------------------------------------



; NOTE:	These functions use LBA28. Maximum visible drive size is 128GiB
;	LBA48 would be needed to access sectors over 128GiB (up to 128PiB)
; -----------------------------------------------------------------------------
; IDEreadsectorsLBA28 -- Read sectors on the hard drive
; IN:	RAX = starting sector to read
;	RCX = number of sectors to read (1 - 256)
;	RDI = memory location to store sectors
;	RBX = hdbase
;	RDX = hdid
; OUT:	RAX = RAX + number of sectors that were read
;	RCX = number of sectors that were read (0 on error)
;	RDI = RDI + (number of sectors * 512)
;	All other registers preserved
IDEreadsectorsLBA28:
	push r10
	push r9
	push rdx
	push rcx
	push rbx
	push rax

	mov r9, rbx		; Store hdbase
	mov r10, rdx		; Store hdid

	push rcx		; Save RCX for use in the read loop
	mov rbx, rcx		; Store number of sectors to read
	cmp rcx, 256
	ja .readsectors_fail	; Over 256? Fail!
	jne .readsectors_skip	; Not 256? No need to modify CL
	xor rcx, rcx		; 0 translates to 256
.readsectors_skip:

	push rax		; Save RAX since we are about to overwrite it
	mov dx, r9w		; e.g.: 0x01F0 [hdbase]
	add dx, 2		; 0x01F2 - Sector count Port 7:0
	mov al, cl		; Read CL sectors
	out dx, al
	pop rax			; Restore RAX which is our sector number
	inc dx			; 0x01F3 - LBA Low Port 7:0
	out dx, al
	inc dx			; 0x01F4 - LBA Mid Port 15:8
	shr rax, 8
	out dx, al
	inc dx			; 0x01F5 - LBA High Port 23:16
	shr rax, 8
	out dx, al
	inc dx			; 0x01F6 - Device Port. Bit 6 set for LBA mode, Bit 4 for device (0 = master, 1 = slave), Bits 3-0 for LBA "Extra High" (27:24)
	shr rax, 8
	and al, 00001111b 	; Clear bits 4-7 just to be safe
	or al, 01000000b	; Turn bit 6 on since we want to use LBA addressing
	or al, r10b		; Set bit 4 for device [hdid]
	out dx, al
	inc dx			; 0x01F7 - Command Port
	mov al, 0x20		; Read sector(s). 0x24 if LBA48
	out dx, al

	mov rcx, 4
.readsectors_wait:
	in al, dx		; Read status from 0x01F7
	test al, 0x80		; BSY flag set?
	jne .readsectors_retry
	test al, 0x08		; DRQ set?
	jne .readsectors_dataready
.readsectors_retry:
	dec rcx
	ja .readsectors_wait
.readsectors_nextsector:
	in al, dx		; Read status from 0x01F7
	test al, 0x80		; BSY flag set?
	jne .readsectors_nextsector
	test al, 0x21		; ERR or DF set?
	jne .readsectors_fail

.readsectors_dataready:
	sub dx, 7		; Data port (0x1F0)
	mov rcx, 256		; Read 
	rep insw		; Copy a 512 byte sector to RDI
	add dx, 7		; Set DX back to status register (0x01F7)
	in al, dx		; Delay ~400ns to allow drive to set new values of BSY and DRQ
	in al, dx
	in al, dx
	in al, dx

.CheckForBusy:
        NOP
        in al, dx
        test al , 0x80		; BSY should be cleared
        jnz .CheckForBusy

	dec rbx			; RBX is the "sectors to read" counter
	cmp rbx, 0
	jne .readsectors_nextsector

	jmp .readsectors_ok

.readsectors_fail:
	pop rcx
	pop rax
	pop rbx
	pop rcx
	pop rdx
	pop r9
	pop r10
	xor rcx, rcx		; Set RCX to 0 since nothing was read
	jmp .end
.readsectors_ok:
	pop rcx
	pop rax
	pop rbx
	add rax, rcx
	pop rcx
	pop rdx
	pop r9
	pop r10
.end:

ret

; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; IDEwritesectorsLBA28 -- Write sectors on the hard drive
; IN:	RAX = starting sector to write
;	RCX = number of sectors to write (1 - 256)
;	RSI = memory location of sectors
;	RBX = hdbase
;	RDX = hdid
; OUT:	RAX = RAX + number of sectors that were written
;	RCX = number of sectors that were written (0 on error)
;	RSI = RSI + (number of sectors * 512)
;	All other registers preserved
IDEwritesectorsLBA28:
	push r10
	push r9
	push rdx
	push rcx
	push rbx
	push rax

	mov r9, rbx		; Store hdbase
	mov r10, rdx		; Store hdid

	push rcx		; Save RCX for use in the write loop
	mov rbx, rcx		; Store number of sectors to write
	cmp rcx, 256
	ja .writesectors_fail	; Over 256? Fail!
	jne .writesectors_skip	; Not 256? No need to modify CL
	xor rcx, rcx		; 0 translates to 256
.writesectors_skip:
	
	push rax		; Save RAX since we are about to overwrite it
	mov dx, r9w		; e.g.: 0x01F0 [hdbase]
	add dx, 2		; 0x01F2 - Sector count Port 7:0
	mov al, cl		; Write CL sectors
	out dx, al
	pop rax			; Restore RAX which is our sector number
	inc dx			; 0x01F3 - LBA Low Port 7:0
	out dx, al
	inc dx			; 0x01F4 - LBA Mid Port 15:8
	shr rax, 8
	out dx, al
	inc dx			; 0x01F5 - LBA High Port 23:16
	shr rax, 8
	out dx, al
	inc dx			; 0x01F6 - Device Port. Bit 6 set for LBA mode, Bit 4 for device (0 = master, 1 = slave), Bits 3-0 for LBA "Extra High" (27:24)
	shr rax, 8		; Bits 7 and 5 are obsolete in LBA mode so set to 0
	and al, 00001111b 	; Clear bits 4-7 just to be safe
	or al, 01000000b	; Turn bit 6 on since we want to use LBA addressing
	or al, r10b		; Set bit 4 for device [hdid]
	out dx, al
	inc dx			; 0x01F7 - Command Port
	mov al, 0x30		; Write sector(s). 0x34 if LBA48
	out dx, al

	mov rcx, 4
.writesectors_wait:
	in al, dx		; Read status from 0x01F7
	test al, 0x80		; BSY flag set?
	jne .writesectors_retry
	test al, 0x08		; DRQ set?
	jne .writesectors_dataready
.writesectors_retry:
	dec rcx
	ja .writesectors_wait
.writesectors_nextsector:
	in al, dx		; Read status from 0x01F7
	test al, 0x80		; BSY flag set?
	jne .writesectors_nextsector
	test al, 0x21		; ERR or DF set?
	jne .writesectors_fail

.writesectors_dataready:
	sub dx, 7		; Data port (0x01F0)
	mov rcx, 256		; Write 256 words (512 bytes)
.writesectors_nextword:
	outsw			; Cannot use rep as a small delay is needed between each out
	sub rcx, 1
	cmp rcx, 0
	jne .writesectors_nextword
	add dx, 7		; Set DX back to Command / Status Register (0x01F7)
	mov al, 0xE7		; Cache Flush command
	out dx, al
	in al, dx		; Delay ~400ns to allow drive to set new values of BSY and DRQ
	in al, dx
	in al, dx
	in al, dx

.CheckForBusy:
        NOP
        in al, dx
        test al , 0x80		; BSY should be cleared
        jnz .CheckForBusy

	dec rbx			; RBX is the "sectors to write" counter
	cmp rbx, 0
	jne .writesectors_nextsector

	jmp .writesectors_ok

.writesectors_fail:
	pop rcx
	pop rax
	pop rbx
	pop rcx
	pop rdx
	pop r9
	pop r10
	xor rcx, rcx		; Set RCX to 0 since nothing was written
	jmp .end
.writesectors_ok:
	pop rcx
	pop rax
	pop rbx
	add rax, rcx
	pop rcx
	pop rdx
	pop r9
	pop r10
.end:

ret


;--------------------------------------------
; IDEIdentifyDrive
; IN: 	r8  = info struc address
;	r9  = hdbase
;	r10 = hdid	
; OUT:	eax = 1 OK, = 0 error
;	info struc filled
;--------------------------------------------
IDEIdentifyDrive:

	push rbx
	push rcx
	push rdx
	push rsi
	push rdi

	xor rax, rax
	xor rdx, rdx

	;SelectTheDrive
	mov dx, r9w				; 0x1f0 = Primary, 0x170 = Secondary
	add dx, 6	 			; DRIVEHEAD offset
	mov eax, 0x000000A0			; DRIVEHEADREG_DEFAULT
	or al, r10b				; 0 = master, 0x10 = slave
	out dx, al
	times 4 nop

	;SendCommand
	mov dx, r9w				; 0x1f0 = Primary, 0x170 = Secondary
	add dx, 7	 			; COMMAND offset
	mov eax, 0x000000EC			; CMD_IDENTIFYDRIVE
	out dx, al
	times 4 nop

;-------
	xor rax, rax
	mov rcx, 4				; num. retry

	;ReadStatus
	mov dx, r9w				; 0x1f0 = Primary, 0x170 = Secondary
	add dx, 7	 			; STATUS offset
.WaitForBSYAndDRQ:				; When the sector of data is ready, the drive sets DRQ and clears BSY
	in al, dx				; Read status

	cmp al, 0x00
	je .fail 				; no device

	test al, 0x80				; BSY flag set?
	jne .retry
	test al, 0x08				; DRQ set?
	jne .dataready
.retry:
	dec rcx
	ja .WaitForBSYAndDRQ


	;ReadStatus
.check_error:
	in al, dx				; Read status again
	test al, 0x80				; BSY flag set?
	jne .check_error
	test al, 0x21				; ERR or DF set?
	jne .fail
;-------
.dataready:

	; Copy Data
	mov dx, r9w				; 0x1f0 = Primary, 0x170 = Secondary
	add dx, 0	 			; DATA offset
	mov rdi , r8
	mov ecx , 256
	rep insw

	mov rax, 1
	jmp .end

.fail:
	mov rax, 0
.end:	
	pop rdi
	pop rsi
	pop rdx
	pop rcx
	pop rbx

ret


;--------------------------------------------
; IDEWaitForDrive
; IN: 	rsi = num. retry
;	r9  = hdbase
;	r10 = hdid
; OUT:	eax = 0 not ready , 1 ready
;--------------------------------------------
IDEWaitForDrive:

	push rbx
	push rcx
	push rdx
	push rsi
	push rdi

	xor rax, rax
	mov rcx, rsi				; num. retry

	;ReadStatus
	mov dx, r9w				; 0x1f0 = Primary, 0x170 = Secondary
	add dx, 7	 			; STATUS offset
.WaitForBSYAndDRQ:				; When the sector of data is ready, the drive sets DRQ and clears BSY
	in al, dx				; Read status

	cmp al, 0x00
	je .fail 				; no device

	test al, 0x80				; BSY flag set?
	jne .retry
	test al, 0x08				; DRQ set?
	jne .ok
.retry:
	dec rcx
	ja .WaitForBSYAndDRQ

	;ReadStatus
.check_error:
	in al, dx				; Read status again
	test al, 0x80				; BSY flag set?
	jne .check_error
	test al, 0x21				; ERR or DF set?
	jne .fail
.ok:

	;SelectTheDrive
	mov dx, r9w				; 0x1f0 = Primary, 0x170 = Secondary
	add dx, 6	 			; DRIVEHEAD offset
	mov eax, 0x000000A0			; DRIVEHEADREG_DEFAULT
	or al, r10b				; 0 = master, 0x10 = slave
	out dx, al
	times 4 nop

	mov rcx, rsi				; num. retry2
	xor rax, rax

	;ReadStatus
	mov dx, r9w				; 0x1f0 = Primary, 0x170 = Secondary
	add dx, 7	 			; STATUS offset
	in al, dx				; Read status
	times 4 nop

      	test al, 0x40 				; status flag DRDY
      	jnz .WaitForBSYAndDRQ2
	jmp .fail


.WaitForBSYAndDRQ2:				; When the sector of data is ready, the drive sets DRQ and clears BSY
	in al, dx				; Read status
	test al, 0x80				; BSY flag set?
	jne .retry2
	test al, 0x08				; DRQ set?
	jne .ok2
.retry2:
	dec rcx
	ja .WaitForBSYAndDRQ2

	;ReadStatus
.check_error2:
	in al, dx				; Read status again
	test al, 0x80				; BSY flag set?
	jne .check_error2
	test al, 0x21				; ERR or DF set?
	jne .fail
.ok2:
	mov rax, 1
	jmp .end

.fail:
	mov rax, 0
.end:
	pop rdi
	pop rsi
	pop rdx
	pop rcx
	pop rbx

ret


;--------------------------------------------------------------------------------------------
; IDEProbe (Based on ATA IDE PIO 0.1 by Vandad Nahavandi Pour)
; IN : none
; OUT: eax
;
; Return value: The return value of the [IDEProbe] function is a BITMASK of different values:
;
;                +--------------------------------------------------------------------+
;                | Value in Hex | Meaning                                             |
;                | ------------ | --------------------------------------------------- |
;                |  0x00000001  | Primary Controller is present.                      |
;                |  0x00000002  | Secondary Controller is present.                    |
;                |  0x00000004  | IDE Drive 0 on the Primary Controller is present.   |
;                |  0x00000008  | IDE Drive 1 on the Primary Controller is present.   |
;                |  0x00000010  | IDE Drive 0 on the Secondary Controller is present. |
;                |  0x00000020  | IDE Drive 1 on the Secondary Controller is present. |
;                +--------------------------------------------------------------------+
IDEProbe:

	push rbx					
	push rcx                                        
	push rdx                                        

	xor ebx, ebx                                    ; Set the temporary result to zero. EAX will be set to EBX at the end of the function's body
	mov edx, 0x000001F3		                ; Write to the HDD Primary IDE Controller LBABITS0TO7 and read the value back later
	mov ecx, 0x00000083                             ; The random value that we will write to each of the controllers and read it back
	mov eax, ecx                                    ; EAX now reads our value to be written to that port and read back from it
	out dx, al                                      ; Send the value to the designated port
	times 0x04 NOP                                  ; Now wait for ~4 clock cycles
	in al, dx                                       ; Read the value back
	cmp eax, ecx                                    ; Is it the same value that we read?
	jne .ProbeTheSecondaryController                ; If no, it means that the primary IDE controller does not exist; probe the secondary controller
	or ebx, 0x00000001	                        ; If yes, indicate in the result that the Primary Controller is present
.ProbeTheSecondaryController:                        	; Now start probing for the secondary IDE controller
	mov eax, ecx                                 	; Get the value again in EAX
	mov edx, 0x00000173		                ; The designated port of the Secondary IDE Controller LBABITS0TO7
	out dx, al	                                ; Put the value in the mentioned port
	times 0x04 NOP                                  ; Wait for ~4 clock cycles before reading the value back
	in al, dx                                    	; Read the value back from the same port
	cmp eax, ecx                                  	; If the value is the same as the one that we wrote to the port, this controller exists too
	jne .CheckForDrivesOnThePrimaryController       ; If no, Start searching for drives on the primary controller
	or ebx, 0x00000002                              ; If yes, then set the corresponding bit in the result to indicate the presense of this controller

    .CheckForDrivesOnThePrimaryController:              ; Start searching for drives on the Primary IDE Controller (PIDEC)
	test ebx, 0x00000001                            ; First see if the bit for the Primary IDE controller is set in the result
	jz .CheckForDrivesOnTheSecondaryController      ; If no, Start probing IDE drives in the secondary controller
      .ProbeDrive0OnPC:                                 ; If yes, start probing IDE drives in the primary controller
	mov eax, (0x000000A0 or 0x00000000)         	; Set the drive/head register with the value for the primary drive
	mov edx, 0x000001F6	                 	; The Drive/Head register of the Primary IDE Controller
	out dx, al                                 	; Probe the designated port and try to read the status register later
	times 0x04 NOP                                 	; Wait for ~4 clock cycles before reading the status register
	mov edx, 0x000001F7                         	; Set the current port to the Status Register
	in al, dx                                  	; Read the Status Register port's value
	test eax, 0x00000040	                 	; Now see if the Drive Ready bit is set?
	jz .ProbeDrive1OnPC                         	; If no, then Drive 0 on the primary IDE controller does not exist
	or ebx, 0x00000004                         	; If yes, then set the corresponding bit in the result
      .ProbeDrive1OnPC:                                 ; Now start probing for Drive 1 on the Primary IDE Controller (PIDEC)
	mov eax, (0x000000A0 or 0x00000010)  	 	; The Secondary Drive bit mask for the Drive/Head Register
	mov edx, 0x000001F6               		; Set the output port to the Drive/Head register
	out dx, al                                  	; Pulse the Drive/Head register and see if Drive 1 on PIDEC exists
	times 0x04 NOP                                  ; Wait for ~4 clock cycles before reading the status register
	mov edx, 0x000001F7                  	 	; Set the input port to the Status Register
	in al, dx                                  	; Read the Status Register ports' value
	test eax, 0x00000040	                 	; See if the Drive Ready bit is set in the Status Register
	jz .CheckForDrivesOnTheSecondaryController  	; If no, start probing for Drive 0 in the Secondary IDE Controller
	or ebx, 0x00000008                 	 	; If yes, set the corresponding bit in the result

    .CheckForDrivesOnTheSecondaryController:            ; Now start searching for available drives on the Secondary IDE Controler (SIDEC)
	test ebx, 0x00000002                           	; First see if the bit for the Secondary IDE controller is set in the result
	jz .EP                                        	; If no, jupm to the end of the procedure
      .ProbeDrive0OnSC:                                 ; If yes, start probing IDE drives in the secondary controller
	mov eax, (0x000000A0 or 0x00000000)         	; Set the drive/head register with the value for the secondary drive
	mov edx, 0x00000176	                 	; The Drive/Head register of the Secondary IDE Controller
	out dx, al                                  	; Probe the designated port and try to read the status register later
	times 0x04 NOP                                 	; Wait for ~4 clock cycles before reading the status register
	mov edx, 0x00000177                         	; Set the current port to the Status Register
	in al, dx                                  	; Read the Status Register port's value
	test eax, 0x00000040                 	 	; Now see if the Drive Ready bit is set?
	jz .ProbeDrive1OnSC                         	; If no, then Drive 0 on the secondary IDE controller does not exist
	or ebx, 0x00000010	                 	; If yes, then set the corresponding bit in the result
      .ProbeDrive1OnSC:                                 ; Now start probing for Drive 1 on the Secondary IDE Controller (PIDEC)
	mov eax, (0x000000A0 or 0x00000010)         	; The Secondary Drive bit mask for the Drive/Head Register
	mov edx, 0x00000176	                 	; Set the output port to the Drive/Head register
	out dx, al                                  	; Pulse the Drive/Head register and see if Drive 1 on PIDEC exists
	times 0x04 NOP                                 	; Wait for ~4 clock cycles before reading the status register
	mov edx, 0x00000177                         	; Set the input port to the Status Register
	in al, dx                                  	; Read the Status Register ports' value
	test eax, 0x00000040                 	 	; See if the Drive Ready bit is set in the Status Register
	jz .EP                                      	; If no, jump to the end of the procedure
	or ebx, 0x00000020	                 	; If yes, set the corresponding bit in the result

    .EP:                                                ; End of the procedure
	xor eax, eax
	mov eax, ebx                                    ; Set the result to the value of the base index register

	pop rdx                                        
	pop rcx                                        
	pop rbx                                       
ret



; -----------------------------------------------------------------------------
; read_cluster -- Read a cluster on the hard drive
; IN:	RAX = starting sector to read
;	RDI = memory location to store cluster (8 sectors)
; OUT:	RAX = RAX + number of sectors that were read
;	RCX = number of sectors that were read (0 on error)
;	RDI = RDI + (number of sectors * 512)
;	All other registers preserved
read_cluster:
	push rbx
	push rdx

	mov rbx, [current.base]
	mov rdx, [current.id]
	mov rcx, 8
	call [current.readsectors]

	pop rdx
	pop rbx
ret

; -----------------------------------------------------------------------------
; write_cluster -- Write a cluster on the hard drive
; IN:	RAX = starting sector to write (8 sectors)
;	RSI = memory location of cluster
; OUT:	RAX = RAX + number of sectors that were written
;	RCX = number of sectors that were written (0 on error)
;	RSI = RSI + (number of sectors * 512)
;	All other registers preserved
write_cluster:
	push rbx
	push rdx

	mov rbx, [current.base]
	mov rdx, [current.id]
	mov rcx, 8
	call [current.writesectors]

	pop rdx
	pop rbx
ret




; =============================================================================
; EOF
