;-----------------------------------------------------------------------------------
; BareNumbersOS v.0.4 - GPL3 - Apr 2013
;
;-----------------------------------------------------------------------------------
; Input Functions
;-----------------------------------------------------------------------------------

;os_input_key_check
;os_input_key_wait
;os_input_string
;os_input_string_with_cursor

; -----------------------------------------------------------------------------
; os_input_key_check -- Scans keyboard for input, but doesn't wait
;  IN:	Nothing
; OUT:	AL = 0 if no key pressed, otherwise  ASCII code
;	AH = 0 if no key pressed, otherwise  scancode
;	Carry flag is set if there was a keystoke, clear if there was not
;	All other registers preserved
os_input_key_check:
	mov ah, [scancode]
	mov al, [key]
	cmp ah, 0
	je os_input_key_check_no_key
	mov byte [key], 0x00		; clear the variable as the keystroke is in AL now
	mov byte [scancode], 0x00	; clear the variable
	stc				; set the carry flag
	ret

os_input_key_check_no_key:
	xor ax, ax
	clc				; clear the carry flag
	ret
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; os_input_key_wait -- Waits for keypress and returns key
;  IN:	Nothing
; OUT:	AL = key pressed
;	AH = scancode
;	All other registers preserved
os_input_key_wait:			; Wait a interrupt
	call os_input_key_check
	jc os_input_key_end		; Key entered...
	hlt				; Halt until another keystroke is received
	jmp os_input_key_wait
os_input_key_end:
	ret
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; os_input_string -- Take string from keyboard entry
;  IN:	RDI = location where string will be stored
;	RCX = number of characters to accept
; OUT:	RCX = length of string that was inputed (NULL not counted)
;	All other registers preserved
os_input_string:
	push rdi
	push rdx			; Counter to keep track of max accepted characters
	push rax

	mov rdx, rcx
	xor rcx, rcx
os_input_string_more:
	call os_input_key_check
	jnc os_input_string_halt	; No key entered... halt until an interrupt is received

	cmp al, 0			; Special key?
	jne os_input_string_char

	cmp ah, 28			; If Enter key pressed, finish
	je os_input_string_done
	cmp ah, 14			; Backspace
	je os_input_string_backspace
	jmp os_input_string_more
	
os_input_string_char:
	cmp al, 32			; In ASCII range (32 - 126)?
	jb os_input_string_more
	cmp al, 126
	ja os_input_string_more
	cmp rcx, rdx			; Check if we have reached the max number of chars
	je os_input_string_more		; Jump if we have (should beep as well)
	stosb				; Store AL at RDI and increment RDI by 1
	inc rcx				; Increment the counter
	call os_print_char		; Display char
	jmp os_input_string_more

os_input_string_backspace:
	cmp rcx, 0			; backspace at the beginning? get a new char
	je os_input_string_more
	call os_dec_cursor		; Decrement the cursor
	mov al, 0x20			; 0x20 is the character for a space
	call os_print_char		; Write over the last typed character with the space
	call os_dec_cursor		; Decremnt the cursor again
	dec rdi				; go back one in the string
	mov byte [rdi], 0x00		; NULL out the char
	dec rcx				; decrement the counter by one
	jmp os_input_string_more

os_input_string_halt:
	hlt				; Halt until another keystroke is received
	jmp os_input_string_more

os_input_string_done:
	mov al, 0x00
	stosb				; We NULL terminate the string

	pop rax
	pop rdx
	pop rdi
	ret
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; os_input_string_with_cursor -- Edit a stored string and show the cursor
;  IN:	RDI = location where string will be stored
;	RCX = number of characters to accept
; OUT:	RCX = length of string that was inputed (NULL not counted)
;	All other registers preserved
os_input_string_with_cursor:

	push rax
	push rdx			; Counter to keep track of max accepted characters
	push rsi
	push rdi

	mov rdx, rcx			; save value
	xor rcx, rcx

	mov rsi, rdi
	call os_print_string		; print the current stored string
	call os_string_length
	add rdi, rcx			; update RDI

.os_input_string_w_more:
	cmp rcx, rdx			; Check if we have reached the max number of chars
	je .last			; Skip the cursor
	mov al, '_'
	call os_print_char		; Show the cursor
	call os_dec_cursor
	.last:

	call os_input_key_wait

	cmp al, 0			; Special key?
	jne .os_input_string_w_char

	cmp ah, 28			; If Enter key pressed, finish
	je .os_input_string_w_done
	cmp ah, 14			; Backspace
	je .os_input_string_w_backspace
	jmp .os_input_string_w_more
	
.os_input_string_w_char:
	cmp al, 32			; In ASCII range (32 - 126)?
	jb .os_input_string_w_more
	cmp al, 126
	ja .os_input_string_w_more
	cmp rcx, rdx			; Check if we have reached the max number of chars
	je .os_input_string_w_more	; Jump if we have (should beep as well)
	stosb				; Store AL at RDI and increment RDI by 1
	inc rcx				; Increment the counter
	call os_print_char		; Display char
	jmp .os_input_string_w_more

.os_input_string_w_backspace:
	cmp rcx, 0			; backspace at the beginning? get a new char
	je .os_input_string_w_more
	cmp rcx, rdx			; Check if we have reached the max number of chars
	je .os_input_string_w_max
	mov al, 0x20			; 0x20 is the character for a space
	call os_print_char		; Write over the cursor with the space
	call os_dec_cursor		; Decrement the cursor
.os_input_string_w_max:
	call os_dec_cursor		; Decrement the cursor (again)

	mov al, 0x20			; 0x20 is the character for a space
	call os_print_char		; Write over the last typed character with the space
	call os_dec_cursor		; Decrement the cursor (again)
	dec rdi				; go back one in the string
	mov byte [rdi], 0x00		; NULL out the char
	dec rcx				; decrement the counter by one
	jmp .os_input_string_w_more

.os_input_string_w_done:
	mov al, 0x00
	stosb				; We NULL terminate the string

	pop rdi
	pop rsi
	pop rdx
	pop rax
	ret
; -----------------------------------------------------------------------------


; =============================================================================
; EOF
