;-----------------------------------------------------------------------------------
; BareNumbersOS v.0.4 - GPL3 - Apr 2013
;
;-----------------------------------------------------------------------------------
; MBR
;-----------------------------------------------------------------------------------
; General x86 Real Mode Memory Map:
;
;    * 0x00000000 - 0x000003FF - Real Mode Interrupt Vector Table
;    * 0x00000400 - 0x000004FF - BIOS Data Area
;    * 0x00000500 - 0x00007BFF - Unused
;    * 0x00007C00 - 0x00007DFF - Our Bootloader
;    * 0x00007E00 - 0x0009FFFF - Unused
;    * 0x000A0000 - 0x000BFFFF - Video RAM (VRAM) Memory
;    * 0x000B0000 - 0x000B7777 - Monochrome Video Memory
;    * 0x000B8000 - 0x000BFFFF - Color Video Memory
;    * 0x000C0000 - 0x000C7FFF - Video ROM BIOS
;    * 0x000C8000 - 0x000EFFFF - BIOS Shadow Area
;    * 0x000F0000 - 0x000FFFFF - System BIOS

USE16
org 0x7C00

	cli
	mov [DriveNumber], dl		; BIOS passes drive number in DL
	xor ax, ax
	mov ss, ax
	mov sp, 0x7C00
	mov si, sp
	push ax
	pop es
	push ax
	pop ds
	sti

; Copy MBR sector to 0x0600 and jump there
	cld
	mov di, 0x0600
	mov cx, 0x0100
	repne movsw
	jmp 0x0000:0x0621
; 0x0621

; Make sure the screen is set to 80x25 color text mode
	mov ax, 0x0003
	int 0x10

; Display boot identifier
  	mov al, [DriveNumber]
  	cmp al, 80h
  	je d1
  	mov al, 'A'
	mov [DriveNumber], 80h		; Forced to prevent some BIOS bugs
  	jmp d2
d1:
  	mov al, 'C'
d2:
  	mov si, DriveLetter
  	mov [si], al
  	call print_string_16               

; Check EBIOS support
	call ebios
	cmp [ebios_present], 1
	jmp ebiosOK
	mov si, msg_NoEbios
	call print_string_16
	jmp $				; Stop
ebiosOK:
	mov si, msg_Ebios
	call print_string_16

; Print message
	mov si, msg_Hello
	call print_string_16

; Check first partition
	mov si, 0x07BE			; 7be-600=1be=446
	cmp byte [si], 0x80
	jne NoActivePartition


; Load SYS64
;-------------------------------
	xor eax, eax
	xor ecx, ecx
	mov bx,  8000h			; Second Stage Address
	mov eax, [StartSYS64]
	mov ecx, [NumSectors]

readsec:
	call readsector
	dec ecx
	cmp ecx, 0
	jne readsec

; Print message
	mov si, msg_2stage
	call print_string_16

	mov dl, [DriveNumber]		; let program know boot drive
	jmp 0x0000:8000h		; Jump to Second Stage

	;jmp $				; Debug


;------------------------------------------------------------------------------
Dap:
	dap_sizepacket 	db 16
	dap_reserved   	db 0
	dap_numsec	dw 1
	dap_bufoff     	dw 0		; seg:off little endian
	dap_bufseg  	dw 0		; 
	dap_startlba1	dd 0		; sector number low 32bit
	dap_startlba2	dd 0		; sector number high 32bit

;------------------------------------------------------------------------------
; Read a sector from a disk, using LBA
; input:	EAX - 32-bit sector number
;		ES:BX - destination buffer
; output:	ES:BX points one byte after the last byte read
;		EAX - next sector
readsector:
	push ecx
	push dx
	push si
	push di

read_it:

	push eax		; Save the sector number

	mov [dap_bufoff], bx
	mov [dap_bufseg], es
	mov [dap_startlba1], eax
	mov si, Dap

	mov dl, [DriveNumber]
	mov ah, 42h		; Extended Read
	int 0x13		; 

	pop eax			; Restore the sector number

	jnc read_ok 		; jump if no error

	push ax
	xor ah, ah		; else, reset and retry
	int 0x13
	pop ax
	jmp read_it

read_ok:
	inc eax 		; next sector
	add bx, 512		; Add bytes per sector
	jnc no_incr_es		; if overflow...

incr_es:
	mov dx, es
	add dh, 0x10		; ...add 1000h to ES
	mov es, dx

no_incr_es:
	pop di
	pop si
	pop dx
	pop ecx
	ret


;------------------------------------------------------------------------------
; 16-bit Function to print a string to the screen
; input:	SI - Address of start of string
;------------------------------------------------------------------------------
print_string_16:			; Output string in SI to screen
	pusha
	mov ah, 0x0E			; int 0x10 teletype function
.repeat:
	lodsb				; Get char from string
	cmp al, 0
	je .done			; If char is zero, end of string
	int 0x10			; Otherwise, print it
	jmp short .repeat
.done:
	popa
	ret

;------------------------------------------------------------------------------
NoActivePartition:
	mov si, msg_NoPartition
	call print_string_16
	jmp $				; Stop

;------------------------------------------------------------------------------
; From Wikipedia:
; INT 13h AH=41h: Check Extensions Present
; Registers(IN):
; AH 	41h = function number for extensions check
; DL 	drive index (e.g. 1st HDD = 80h)
; BX 	55AAh
; Results(OUT):
; CF 	Set On Not Present, Clear If Present
; AH 	Error Code or Major Version Number
; BX 	AA55h
; CX 	Interface support bitmask:
;	1 - Device Access using the packet structure
;	2 - Drive Locking and Ejecting
;	4 - Enhanced Disk Drive Support (EDD)
;------------------------------------------------------------------------------
; Check if EBIOS is supported
; ebios_present : 0 if not supported, 1 if supported
;------------------------------------------------------------------------------
ebios:
    mov     [ebios_present], 0      ; clear EBIOS support flag
    mov     ah, 0x41                ; function 0x41
    mov     bx, 0x55AA              ; check signature
    int     0x13
    jc      .ebios_check_done
    cmp     bx, 0xAA55              ; check BX = 0xAA55
    jne     .ebios_check_done
    test    cl, 0x01                ; check packet structure support
    setnz   [ebios_present]         ; EBIOS supported, set flag
.ebios_check_done:

ret

;------------------------------------------------------------------------
msg_Hello       db 13,10,"BareNumbersOS",13,10,"Loading.",0
msg_2stage      db "..",13,10,0
msg_NoPartition db "No Active Partition",13,10,0
msg_NoEbios	db " No EBios",13,10,0
msg_Ebios	db " EBios OK",13,10,0
ebios_present   db 0   			; 1 if EBIOS is supported, 0 otherwise.
DriveLetter     db "0",0
DriveNumber 	db 0x00
StartSYS64  	dd 40			; first LBA sector
NumSectors  	dd 32*2 		; Size of SYS64  ( in sectors )

times 446-$+$$ db 0
 
;tables db "XXXXXXXXXXXXXXXX  DO NOT OVERWRITE THIS AREA!!! XXXXXXXXXXXXXXXX" ; 64 bytes in length
part1status db 80h			; 80h=active 0h=not bootable  - 0x07BE
part1CHSs   db 0,0,0
part1type   db 0xBA			; BareNumbersOS Tag
part1CHSe   db 0,0,0
part1LBAs   dd 40			; Start Partition
part1nsec   dd 16384			; Size Partition (little endian) - 32 bit -> max 2 TiB

part2status db 0h			; 80h=active 0h=not bootable
part2CHSs   db 0,0,0
part2type   db 0
part2CHSe   db 0,0,0
part2LBAs   dd 0
part2nsec   dd 0			; little endian

part3status db 0h			; 80h=active 0h=not bootable
part3CHSs   db 0,0,0
part3type   db 0
part3CHSe   db 0,0,0
part3LBAs   dd 0
part3nsec   dd 0			; little endian

part4status db 0h			; 80h=active 0h=not bootable
part4CHSs   db 0,0,0
part4type   db 0
part4CHSe   db 0,0,0
part4LBAs   dd 0
part4nsec   dd 0			; little endian

sign dw 0xAA55				; 0x07FE (MBR relocated to 0x0600)
